# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Internal dictionary. '''

# pylint: disable=unused-import
# ruff: noqa: F401


# TODO: Consider a dictionary factory to allow 'mutables' closure
#       to be referenced in the '__setitem__' and '__delitem__' methods.


from __future__ import annotations

from . import imports as __
from . import immutables as _immutables


_H = __.typx.TypeVar( '_H' )
_V = __.typx.TypeVar( '_V' )


class AccretiveDictionary(
    _immutables.ConcealerExtension,
    dict[ _H, _V ],
    __.typx.Generic[ _H, _V ],
):
    ''' Accretive subclass of :py:class:`dict`.

        Can be used as an instance dictionary.

        Prevents attempts to mutate dictionary via inherited interface.
    '''

    def __init__(
        self,
        *iterables: __.DictionaryPositionalArgument[ _H, _V ],
        **entries: __.DictionaryNominativeArgument[ _V ],
    ):
        super( ).__init__( )
        self.update( *iterables, **entries )

    def __delitem__( self, key: _H ) -> None:
        from .exceptions import EntryImmutabilityError
        raise EntryImmutabilityError( key )

    def __setitem__( self, key: _H, value: _V ) -> None:
        from .exceptions import EntryImmutabilityError
        if key in self: raise EntryImmutabilityError( key )
        super( ).__setitem__( key, value )

    def clear( self ) -> __.typx.Never:
        ''' Raises exception. Cannot clear immutable entries. '''
        from .exceptions import OperationInvalidity
        raise OperationInvalidity( 'clear' )

    def copy( self ) -> __.typx.Self:
        ''' Provides fresh copy of dictionary. '''
        return type( self )( self )

    def pop( # pylint: disable=unused-argument
        self, key: _H, default: __.Absential[ _V ] = __.absent
    ) -> __.typx.Never:
        ''' Raises exception. Cannot pop immutable entry. '''
        from .exceptions import OperationInvalidity
        raise OperationInvalidity( 'pop' )

    def popitem( self ) -> __.typx.Never:
        ''' Raises exception. Cannot pop immutable entry. '''
        from .exceptions import OperationInvalidity
        raise OperationInvalidity( 'popitem' )

    def update( # type: ignore
        self,
        *iterables: __.DictionaryPositionalArgument[ _H, _V ],
        **entries: __.DictionaryNominativeArgument[ _V ],
    ) -> None:
        ''' Adds new entries as a batch. '''
        from itertools import chain
        # Add values in order received, enforcing no alteration.
        for indicator, value in chain.from_iterable( map( # type: ignore
            lambda element: ( # pyright: ignore
                element.items( )
                if isinstance( element, __.cabc.Mapping )
                else element
            ),
            ( *iterables, entries )
        ) ): self[ indicator ] = value # type: ignore



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Accretive data structures. '''


from . import __
from . import qaliases
# --- BEGIN: Injected by Copier ---
from . import exceptions
# --- END: Injected by Copier ---

from .classes import *
from .dictionaries import *
from .modules import *
from .namespaces import *
from .objects import *


__version__ = '3.0'


_attribute_visibility_includes_ = frozenset( ( '__version__', ) )
__.reclassify_modules( __name__, recursive = True )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


# ruff: noqa: F811


# pylint: disable=line-too-long
''' Accretive objects.

    Provides the base class for objects with accretive attributes. Once an
    attribute is set on an instance, it cannot be reassigned or deleted.

    The implementation uses a special dictionary type for attribute storage
    that enforces the accretive behavior. This makes it suitable as a base
    class for:

    * Configuration objects
    * Plugin interfaces
    * Immutable data containers
    * Objects requiring attribute stability

    >>> from accretive import Object
    >>> obj = Object( )
    >>> obj.x = 1  # Add new instance attribute
    >>> obj.y = 2  # Add another instance attribute
    >>> obj.x = 3  # Attempt modification
    Traceback (most recent call last):
        ...
    accretive.exceptions.AttributeImmutabilityError: Cannot reassign or delete existing attribute 'x'.

    The `accretive` decorator can be used to make any class accretive:

    >>> from accretive import accretive
    >>> @accretive
    ... class Config:
    ...     def __init__( self, debug = False ):
    ...         self.debug = debug
    ...
    >>> config = Config( debug = True )
    >>> config.debug  # Access existing attribute
    True
    >>> config.verbose = True  # Add new attribute
    >>> config.debug = False  # Attempt to modify existing attribute
    Traceback (most recent call last):
        ...
    accretive.exceptions.AttributeImmutabilityError: Cannot reassign or delete existing attribute 'debug'.
'''
# pylint: enable=line-too-long


from . import __


_behavior = 'accretion'


def _check_behavior( obj: object ) -> bool:
    behaviors: __.cabc.MutableSet[ str ]
    if _check_dict( obj ):
        attributes = getattr( obj, '__dict__' )
        behaviors = attributes.get( '_behaviors_', set( ) )
    else: behaviors = getattr( obj, '_behaviors_', set( ) )
    return _behavior in behaviors


def _check_dict( obj: object ) -> bool:
    # Return False even if '__dict__' in '__slots__'.
    if hasattr( obj, '__slots__' ): return False
    return hasattr( obj, '__dict__' )


@__.typx.overload
def accretive( # pragma: no branch
    class_: type[ __.C ], *,
    docstring: __.Absential[ __.typx.Optional[ str ] ] = __.absent,
    mutables: __.cabc.Collection[ str ] = ( )
) -> type[ __.C ]: ...


@__.typx.overload
def accretive( # pragma: no branch
    class_: __.AbsentSingleton, *,
    docstring: __.Absential[ __.typx.Optional[ str ] ] = __.absent,
    mutables: __.cabc.Collection[ str ] = ( )
) -> __.typx.Callable[ [ type[ __.C ] ], type[ __.C ] ]: ...


def accretive( # pylint: disable=too-complex,too-many-statements
    class_: __.Absential[ type[ __.C ] ] = __.absent, *,
    docstring: __.Absential[ __.typx.Optional[ str ] ] = __.absent,
    mutables: __.cabc.Collection[ str ] = ( )
) -> __.typx.Union[
    type[ __.C ], __.typx.Callable[ [ type[ __.C ] ], type[ __.C ] ]
]:
    ''' Decorator which makes class accretive after initialization.

        Cannot be applied to classes which define their own __setattr__
        or __delattr__ methods.

        This decorator can be used in different ways:

        1. Simple decorator:

           >>> @accretive
           ... class Config:
           ...     pass

        2. With parameters:

           >>> @accretive( mutables = ( 'version', ) )
           ... class Config:
           ...     pass
    '''
    def decorator( cls: type[ __.C ] ) -> type[ __.C ]: # pylint: disable=too-many-statements
        if not __.is_absent( docstring ): cls.__doc__ = docstring
        for method in ( '__setattr__', '__delattr__' ):
            if method in cls.__dict__:
                from .exceptions import DecoratorCompatibilityError
                raise DecoratorCompatibilityError( cls.__name__, method )
        original_init = next(
            base.__dict__[ '__init__' ] for base in cls.__mro__
            if '__init__' in base.__dict__ ) # pylint: disable=magic-value-comparison
        mutables_ = frozenset( mutables )

        def __init__(
            self: object, *posargs: __.typx.Any, **nomargs: __.typx.Any
        ) -> None:
            original_init( self, *posargs, **nomargs )
            behaviors: __.cabc.MutableSet[ str ]
            if _check_dict( self ):
                attributes = getattr( self, '__dict__' )
                behaviors = attributes.get( '_behaviors_', set( ) )
                if not behaviors: attributes[ '_behaviors_' ] = behaviors
            else:
                behaviors = getattr( self, '_behaviors_', set( ) )
                if not behaviors: setattr( self, '_behaviors_', behaviors )
            behaviors.add( _behavior )

        def __delattr__( self: object, name: str ) -> None:
            if name in mutables_:
                super( cls, self ).__delattr__( name )
                return
            if _check_behavior( self ): # pragma: no branch
                from .exceptions import AttributeImmutabilityError
                raise AttributeImmutabilityError( name )
            super( cls, self ).__delattr__( name ) # pragma: no cover

        def __setattr__( self: object, name: str, value: __.typx.Any ) -> None:
            if name in mutables_:
                super( cls, self ).__setattr__( name, value )
                return
            if _check_behavior( self ) and hasattr( self, name ):
                from .exceptions import AttributeImmutabilityError
                raise AttributeImmutabilityError( name )
            super( cls, self ).__setattr__( name, value )

        cls.__init__ = __init__
        cls.__delattr__ = __delattr__
        cls.__setattr__ = __setattr__
        return cls

    if not __.is_absent( class_ ): return decorator( class_ )
    return decorator # No class to decorate; keyword arguments only.


@accretive
class Object:
    ''' Accretive objects. '''

    __slots__ = ( '__dict__', '_behaviors_' )

    def __repr__( self ) -> str:
        return "{fqname}( )".format( fqname = __.calculate_fqname( self ) )

Object.__doc__ = __.generate_docstring(
    Object, 'instance attributes accretion' )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


# pylint: disable=line-too-long
''' Accretive classes.

    Provides metaclasses for creating classes with accretive attributes. Once a
    class attribute is set, it cannot be reassigned or deleted.

    The implementation includes:

    * ``Class``: Standard metaclass for accretive classes; derived from
      :py:class:`type`.
    * ``ABCFactory``: Metaclass for abstract base classes; derived from
      :py:class:`abc.ABCMeta`.
    * ``ProtocolClass``: Metaclass for protocol classes; derived from
      :py:class:`typing.Protocol`.

    These metaclasses are particularly useful for:

    * Creating classes with constant class attributes
    * Defining stable abstract base classes
    * Building protocol classes with fixed interfaces

    >>> from accretive import Class
    >>> class Example( metaclass = Class ):
    ...     x = 1
    >>> Example.y = 2  # Add new class attribute
    >>> Example.x = 3  # Attempt reassignment
    Traceback (most recent call last):
        ...
    accretive.exceptions.AttributeImmutabilityError: Cannot reassign or delete attribute 'x'.

    For cases where some attributes need to remain mutable, use the ``mutables`` parameter:

    >>> class Config( metaclass = Class, mutables = ( 'version', ) ):
    ...     name = 'MyApp'
    ...     version = '1.0.0'
    >>> Config.version = '1.0.1'  # Can modify designated mutable attributes
    >>> Config.version
    '1.0.1'
    >>> Config.name = 'NewApp'    # Other attributes remain immutable
    Traceback (most recent call last):
        ...
    accretive.exceptions.AttributeImmutabilityError: Cannot reassign or delete attribute 'name'.
'''
# pylint: enable=line-too-long

# TODO? Allow predicate functions and regex patterns as mutability checkers.


from __future__ import annotations

from . import __


ClassDecorators: __.typx.TypeAlias = (
    __.cabc.Iterable[ __.cabc.Callable[ [ type ], type ] ] )


_behavior = 'accretion'


class Class( type ):
    ''' Accretive class factory. '''

    def __new__( # pylint: disable=too-many-arguments
        clscls: type[ Class ],
        name: str,
        bases: tuple[ type, ... ],
        namespace: dict[ str, __.typx.Any ], *,
        decorators: ClassDecorators = ( ),
        docstring: __.Absential[ __.typx.Optional[ str ] ] = __.absent,
        mutables: __.cabc.Collection[ str ] = ( ),
        **args: __.typx.Any
    ) -> Class:
        class_ = type.__new__(
            clscls, name, bases, namespace, **args )
        return _class__new__(
            class_,
            decorators = decorators,
            docstring = docstring,
            mutables = mutables )

    def __init__( selfclass, *posargs: __.typx.Any, **nomargs: __.typx.Any ):
        super( ).__init__( *posargs, **nomargs )
        _class__init__( selfclass )

    def __delattr__( selfclass, name: str ) -> None:
        if not _class__delattr__( selfclass, name ):
            super( ).__delattr__( name )

    def __setattr__( selfclass, name: str, value: __.typx.Any ) -> None:
        if not _class__setattr__( selfclass, name ):
            super( ).__setattr__( name, value )

Class.__doc__ = __.generate_docstring(
    Class,
    'description of class factory class',
    'class attributes accretion' )


@__.typx.dataclass_transform( kw_only_default = True )
class Dataclass( Class ):
    ''' Accretive dataclass factory. '''

    def __new__( # pylint: disable=too-many-arguments
        clscls: type[ Dataclass ],
        name: str,
        bases: tuple[ type, ... ],
        namespace: dict[ str, __.typx.Any ], *,
        decorators: ClassDecorators = ( ),
        docstring: __.Absential[ __.typx.Optional[ str ] ] = __.absent,
        mutables: __.cabc.Collection[ str ] = ( ),
        **args: __.typx.Any
    ) -> Dataclass:
        decorators_ = (
            __.dcls.dataclass( kw_only = True, slots = True ),
            *decorators )
        return Class.__new__( # pyright: ignore
            clscls, name, bases, namespace,
            decorators = decorators_,
            docstring = docstring,
            mutables = mutables,
            **args )

Dataclass.__doc__ = __.generate_docstring(
    Dataclass,
    'description of class factory class',
    'class attributes accretion' )


@__.typx.dataclass_transform( frozen_default = True, kw_only_default = True )
class CompleteDataclass( Class ):
    ''' Accretive dataclass factory.

        Dataclasses from this factory produce immutable instances. '''
    def __new__( # pylint: disable=too-many-arguments
        clscls: type[ CompleteDataclass ],
        name: str,
        bases: tuple[ type, ... ],
        namespace: dict[ str, __.typx.Any ], *,
        decorators: ClassDecorators = ( ),
        docstring: __.Absential[ __.typx.Optional[ str ] ] = __.absent,
        mutables: __.cabc.Collection[ str ] = ( ),
        **args: __.typx.Any
    ) -> CompleteDataclass:
        decorators_ = (
            __.dcls.dataclass( frozen = True, kw_only = True, slots = True ),
            *decorators )
        return Class.__new__( # pyright: ignore
            clscls, name, bases, namespace,
            decorators = decorators_,
            docstring = docstring,
            mutables = mutables,
            **args )

CompleteDataclass.__doc__ = __.generate_docstring(
    CompleteDataclass,
    'description of class factory class',
    'class attributes accretion' )


class ABCFactory( __.abc.ABCMeta ):
    ''' Accretive abstract base class factory. '''

    def __new__( # pylint: disable=too-many-arguments
        clscls: type[ ABCFactory ],
        name: str,
        bases: tuple[ type, ... ],
        namespace: dict[ str, __.typx.Any ], *,
        decorators: ClassDecorators = ( ),
        docstring: __.Absential[ __.typx.Optional[ str ] ] = __.absent,
        mutables: __.cabc.Collection[ str ] = ( ),
        **args: __.typx.Any
    ) -> ABCFactory:
        class_ = __.abc.ABCMeta.__new__(
            clscls, name, bases, namespace, **args )
        return _class__new__(
            class_,
            decorators = decorators,
            docstring = docstring,
            mutables = mutables )

    def __init__( selfclass, *posargs: __.typx.Any, **nomargs: __.typx.Any ):
        super( ).__init__( *posargs, **nomargs )
        _class__init__( selfclass )

    def __delattr__( selfclass, name: str ) -> None:
        if not _class__delattr__( selfclass, name ):
            super( ).__delattr__( name )

    def __setattr__( selfclass, name: str, value: __.typx.Any ) -> None:
        if not _class__setattr__( selfclass, name ):
            super( ).__setattr__( name, value )

ABCFactory.__doc__ = __.generate_docstring(
    ABCFactory,
    'description of class factory class',
    'class attributes accretion' )


# pylint: disable=bad-classmethod-argument,no-self-argument
class ProtocolClass( type( __.typx.Protocol ) ):
    ''' Accretive protocol class factory. '''

    def __new__( # pylint: disable=too-many-arguments
        clscls: type[ ProtocolClass ],
        name: str,
        bases: tuple[ type, ... ],
        namespace: dict[ str, __.typx.Any ], *,
        decorators: ClassDecorators = ( ),
        docstring: __.Absential[ __.typx.Optional[ str ] ] = __.absent,
        mutables: __.cabc.Collection[ str ] = ( ),
        **args: __.typx.Any
    ) -> ProtocolClass:
        class_ = super( ProtocolClass, clscls ).__new__( # pylint: disable=too-many-function-args
            clscls, name, bases, namespace, **args )
        return _class__new__(
            class_,
            decorators = decorators,
            docstring = docstring,
            mutables = mutables )

    def __init__( selfclass, *posargs: __.typx.Any, **nomargs: __.typx.Any ):
        super( ).__init__( *posargs, **nomargs )
        _class__init__( selfclass )

    def __delattr__( selfclass, name: str ) -> None:
        if not _class__delattr__( selfclass, name ):
            super( ).__delattr__( name )

    def __setattr__( selfclass, name: str, value: __.typx.Any ) -> None:
        if not _class__setattr__( selfclass, name ):
            super( ).__setattr__( name, value )
# pylint: enable=bad-classmethod-argument,no-self-argument

ProtocolClass.__doc__ = __.generate_docstring(
    ProtocolClass,
    'description of class factory class',
    'class attributes accretion' )


# pylint: disable=bad-classmethod-argument,no-self-argument
@__.typx.dataclass_transform( kw_only_default = True )
class ProtocolDataclass( ProtocolClass ):
    ''' Accretive protocol dataclass factory. '''
    def __new__( # pylint: disable=too-many-arguments
        clscls: type[ ProtocolDataclass ],
        name: str,
        bases: tuple[ type, ... ],
        namespace: dict[ str, __.typx.Any ], *,
        decorators: ClassDecorators = ( ),
        docstring: __.Absential[ __.typx.Optional[ str ] ] = __.absent,
        mutables: __.cabc.Collection[ str ] = ( ),
        **args: __.typx.Any
    ) -> ProtocolDataclass:
        decorators_ = (
            __.dcls.dataclass( kw_only = True, slots = True ),
            *decorators )
        return ProtocolClass.__new__( # pyright: ignore
            clscls, name, bases, namespace,
            decorators = decorators_,
            docstring = docstring,
            mutables = mutables,
            **args )
# pylint: enable=bad-classmethod-argument,no-self-argument

ProtocolDataclass.__doc__ = __.generate_docstring(
    ProtocolDataclass,
    'description of class factory class',
    'class attributes accretion' )


# pylint: disable=bad-classmethod-argument,no-self-argument
@__.typx.dataclass_transform( frozen_default = True, kw_only_default = True )
class CompleteProtocolDataclass( ProtocolClass ):
    ''' Accretive protocol dataclass factory.

        Dataclasses from this factory produce immutable instances. '''
    def __new__( # pylint: disable=too-many-arguments
        clscls: type[ CompleteProtocolDataclass ],
        name: str,
        bases: tuple[ type, ... ],
        namespace: dict[ str, __.typx.Any ], *,
        decorators: ClassDecorators = ( ),
        docstring: __.Absential[ __.typx.Optional[ str ] ] = __.absent,
        mutables: __.cabc.Collection[ str ] = ( ),
        **args: __.typx.Any
    ) -> CompleteProtocolDataclass:
        decorators_ = (
            __.dcls.dataclass( frozen = True, kw_only = True, slots = True ),
            *decorators )
        return ProtocolClass.__new__( # pyright: ignore
            clscls, name, bases, namespace,
            decorators = decorators_,
            docstring = docstring,
            mutables = mutables,
            **args )
# pylint: enable=bad-classmethod-argument,no-self-argument

CompleteProtocolDataclass.__doc__ = __.generate_docstring(
    CompleteProtocolDataclass,
    'description of class factory class',
    'class attributes accretion' )


def _accumulate_mutables(
    class_: type, mutables: __.cabc.Collection[ str ]
) -> frozenset[ str ]:
    return frozenset( mutables ).union( *(
        frozenset( base.__dict__.get( '_class_mutables_', ( ) ) )
        for base in class_.__mro__ ) )

# pylint: disable=protected-access
def _class__new__(
    original: type,
    decorators: ClassDecorators = ( ),
    docstring: __.Absential[ __.typx.Optional[ str ] ] = __.absent,
    mutables: __.cabc.Collection[ str ] = ( ),
) -> type:
    # Some decorators create new classes, which invokes this method again.
    # Short-circuit to prevent recursive decoration and other tangles.
    class_decorators_ = original.__dict__.get( '_class_decorators_', [ ] )
    if class_decorators_: return original
    if not __.is_absent( docstring ): original.__doc__ = docstring
    original._class_mutables_ = _accumulate_mutables( original, mutables )
    original._class_decorators_ = class_decorators_
    reproduction = original
    for decorator in decorators:
        class_decorators_.append( decorator )
        reproduction = decorator( original )
        if original is not reproduction:
            __.repair_class_reproduction( original, reproduction )
        original = reproduction
    class_decorators_.clear( )  # Flag '__init__' to enable accretion
    return reproduction
# pylint: enable=protected-access


# pylint: disable=protected-access
def _class__init__( class_: type ) -> None:
    # Some metaclasses add class attributes in '__init__' method.
    # So, we wait until last possible moment to set immutability.
    # Consult class attributes dictionary to ignore immutable base classes.
    cdict = class_.__dict__
    if cdict.get( '_class_decorators_' ): return
    del class_._class_decorators_
    if ( class_behaviors := cdict.get( '_class_behaviors_' ) ):
        class_behaviors.add( _behavior )
    else: class_._class_behaviors_ = { _behavior }
# pylint: enable=protected-access


def _class__delattr__( class_: type, name: str ) -> bool:
    # Consult class attributes dictionary to ignore accretive base classes.
    cdict = class_.__dict__
    if name in cdict.get( '_class_mutables_', ( ) ): return False
    if _behavior not in cdict.get( '_class_behaviors_', ( ) ): return False
    from .exceptions import AttributeImmutabilityError
    raise AttributeImmutabilityError( name )


def _class__setattr__( class_: type, name: str ) -> bool:
    # Consult class attributes dictionary to ignore accretive base classes.
    cdict = class_.__dict__
    if name in cdict.get( '_class_mutables_', ( ) ): return False
    if _behavior not in cdict.get( '_class_behaviors_', ( ) ): return False
    if hasattr( class_, name ):
        from .exceptions import AttributeImmutabilityError
        raise AttributeImmutabilityError( name )
    return False  # Allow setting new attributes



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Family of exceptions for package API.

    Provides a hierarchy of exceptions that are raised when accretive behavior
    is violated. The hierarchy is designed to allow both specific and general
    exception handling.
'''


from . import __ # pylint: disable=cyclic-import


class Omniexception( __.ImmutableObject, BaseException ):
    ''' Base for all exceptions raised by package API. '''

    _attribute_visibility_includes_: __.cabc.Collection[ str ] = (
        frozenset( ( '__cause__', '__context__', ) ) )


class Omnierror( Omniexception, Exception ):
    ''' Base for error exceptions raised by package API. '''


class AttributeImmutabilityError( Omnierror, AttributeError, TypeError ):
    ''' Attempt to reassign or delete immutable attribute. '''

    def __init__( self, name: str ) -> None:
        super( ).__init__( f"Cannot reassign or delete attribute {name!r}." )


class EntryImmutabilityError( Omnierror, TypeError ):
    ''' Attempt to update or remove immutable dictionary entry. '''

    def __init__( self, indicator: __.cabc.Hashable ) -> None:
        super( ).__init__(
            f"Cannot alter or remove existing entry for {indicator!r}." )


class EntryValidityError( Omnierror, ValueError ):
    ''' Attempt to add invalid entry to dictionary. '''

    def __init__(
        self, indicator: __.cabc.Hashable, value: __.typx.Any
    ) -> None:
        super( ).__init__(
            f"Cannot add invalid entry with key, {indicator!r}, "
            f"and value, {value!r}, to dictionary." )


class DecoratorCompatibilityError( Omnierror, TypeError ):
    ''' Attempt to apply decorator to incompatible class. '''

    def __init__( self, class_name: str, method_name: str ) -> None:
        super( ).__init__(
            f"Cannot apply accretive decorator to {class_name!r} "
            f"because it defines {method_name!r}.")



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Common imports and type aliases used throughout the package. '''

# pylint: disable=unused-import
# ruff: noqa: F401


from __future__ import annotations

import abc
import collections.abc as cabc
import dataclasses as dcls
import types

import typing_extensions as typx
# --- BEGIN: Injected by Copier ---

# --- END: Injected by Copier ---


from absence import AbsentSingleton, Absential, absent, is_absent


C = typx.TypeVar( 'C' )  # Class
H = typx.TypeVar( 'H', bound = cabc.Hashable ) # Hash Key
V = typx.TypeVar( 'V' ) # Value


ClassDecorators: typx.TypeAlias = (
    cabc.Iterable[ cabc.Callable[ [ type ], type ] ] )
ComparisonResult: typx.TypeAlias = bool | types.NotImplementedType
DictionaryNominativeArgument: typx.TypeAlias = typx.Annotated[
    V,
    typx.Doc(
        'Zero or more keyword arguments from which to initialize '
        'dictionary data.' ),
]
DictionaryPositionalArgument: typx.TypeAlias = typx.Annotated[
    cabc.Mapping[ H, V ] | cabc.Iterable[ tuple[ H, V ] ],
    typx.Doc(
        'Zero or more iterables from which to initialize dictionary data. '
        'Each iterable must be dictionary or sequence of key-value pairs. '
        'Duplicate keys will result in an error.' ),
]
DictionaryProducer: typx.TypeAlias = typx.Annotated[
    cabc.Callable[ [ ], V ],
    typx.Doc(
        'Callable which produces values for absent dictionary entries.' ),
]
DictionaryValidator: typx.TypeAlias = typx.Annotated[
    cabc.Callable[ [ H, V ], bool ],
    typx.Doc(
        'Callable which validates entries before addition to dictionary.' ),
]
ModuleReclassifier: typx.TypeAlias = cabc.Callable[
    [ cabc.Mapping[ str, typx.Any ] ], None ]



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Accretive modules.

    Provides a module type that enforces attribute immutability after
    assignment. This helps ensure that module-level constants remain constant
    and that module interfaces remain stable during runtime.

    The module implementation is derived from :py:class:`types.ModuleType` and
    adds accretive behavior. This makes it particularly useful for:

    * Ensuring constants remain constant
    * Preventing accidental modification of module interfaces

    Also provides a convenience function:

    * ``reclassify_modules``: Converts existing modules to accretive modules.
'''


from . import __


class Module( __.types.ModuleType ):
    ''' Accretive modules. '''

    def __delattr__( self, name: str ) -> None:
        from .exceptions import AttributeImmutabilityError
        raise AttributeImmutabilityError( name )

    def __setattr__( self, name: str, value: __.typx.Any ) -> None:
        if hasattr( self, name ):
            from .exceptions import AttributeImmutabilityError
            raise AttributeImmutabilityError( name )
        super( ).__setattr__( name, value )

Module.__doc__ = __.generate_docstring(
    Module, 'description of module', 'module attributes accretion' )


def reclassify_modules(
    attributes: __.typx.Annotated[
        __.cabc.Mapping[ str, __.typx.Any ] | __.types.ModuleType | str,
        __.typx.Doc(
            'Module, module name, or dictionary of object attributes.' ),
    ],
    recursive: __.typx.Annotated[
        bool, __.typx.Doc( 'Recursively reclassify package modules?' ),
    ] = False,
) -> None:
    ''' Reclassifies modules to be accretive.

        Can operate on individual modules or entire package hierarchies.

        Notes
        -----
        * Only converts modules within the same package to prevent unintended
          modifications to external modules.
        * When used with a dictionary, converts any module objects found as
          values if they belong to the same package.
        * Has no effect on already-accretive modules.
    '''
    from inspect import ismodule
    from sys import modules
    if isinstance( attributes, str ):
        attributes = modules[ attributes ]
    if isinstance( attributes, __.types.ModuleType ):
        module = attributes
        attributes = attributes.__dict__
    else: module = None
    package_name = (
        attributes.get( '__package__' ) or attributes.get( '__name__' ) )
    if not package_name: return
    for value in attributes.values( ):
        if not ismodule( value ): continue
        if not value.__name__.startswith( f"{package_name}." ): continue
        if recursive: reclassify_modules( value, recursive = True )
        if isinstance( value, Module ): continue
        value.__class__ = Module
    if module and not isinstance( module, Module ):
        module.__class__ = Module



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


# pylint: disable=line-too-long
''' Accretive dictionaries.

    Dictionaries which can grow but never shrink. Once an entry is added, it
    cannot be modified or removed.

    * :py:class:`AbstractDictionary`:
      Base class defining the accretive dictionary interface. Implementations
      must provide ``__getitem__``, ``__iter__``, ``__len__``, and storage
      methods.

    * :py:class:`Dictionary`:
      Standard implementation of an accretive dictionary. Supports all usual
      dict operations except those that would modify or remove existing
      entries.

    * :py:class:`ProducerDictionary`:
      Automatically generates values for missing keys using a supplied factory
      function. Similar to :py:class:`collections.defaultdict` but with
      accretive behavior.

    * :py:class:`ValidatorDictionary`:
      Validates entries before addition using a supplied predicate function.

    * :py:class:`ProducerValidatorDictionary`:
      Combines producer and validator behaviors. Generated values must pass
      validation before being added.

    >>> from accretive import Dictionary
    >>> d = Dictionary( apples = 12, bananas = 6 )
    >>> d[ 'cherries' ] = 42  # Add new entry
    >>> d[ 'apples' ] = 14    # Attempt modification
    Traceback (most recent call last):
        ...
    accretive.exceptions.EntryImmutabilityError: Cannot alter or remove existing entry for 'apples'.
    >>> del d[ 'bananas' ]    # Attempt removal
    Traceback (most recent call last):
        ...
    accretive.exceptions.EntryImmutabilityError: Cannot alter or remove existing entry for 'bananas'.

    >>> from accretive import ProducerDictionary
    >>> d = ProducerDictionary( list )  # list() called for missing keys
    >>> d[ 'new' ]
    []
    >>> d[ 'new' ].append( 1 )  # List is mutable, but entry is fixed
    >>> d[ 'new' ] = [ ]  # Attempt modification
    Traceback (most recent call last):
        ...
    accretive.exceptions.EntryImmutabilityError: Cannot alter or remove existing entry for 'new'.

    >>> from accretive import ValidatorDictionary
    >>> d = ValidatorDictionary( lambda k, v: isinstance( v, int ) )
    >>> d[ 'valid' ] = 42  # Passes validation
    >>> d[ 'invalid' ] = 'str'  # Fails validation
    Traceback (most recent call last):
        ...
    accretive.exceptions.EntryValidityError: Cannot add invalid entry with key, 'invalid', and value, 'str', to dictionary.
'''
# pylint: enable=line-too-long


from . import __
from . import classes as _classes


class AbstractDictionary( __.cabc.Mapping[ __.H, __.V ] ):
    ''' Abstract base class for dictionaries that can grow but not shrink.

        An accretive dictionary allows new entries to be added but prevents
        modification or removal of existing entries. This provides a middle
        ground between immutable and fully mutable mappings.

        Implementations must provide:
        - __getitem__, __iter__, __len__
        - _pre_setitem_ for entry validation/preparation
        - _store_item_ for storage implementation
    '''

    @__.abc.abstractmethod
    def __iter__( self ) -> __.cabc.Iterator[ __.H ]:
        raise NotImplementedError # pragma: no coverage

    @__.abc.abstractmethod
    def __len__( self ) -> int:
        raise NotImplementedError # pragma: no coverage

    @__.abc.abstractmethod
    def __getitem__( self, key: __.H ) -> __.V:
        raise NotImplementedError # pragma: no coverage

    def _pre_setitem_( # pylint: disable=no-self-use
        self, key: __.H, value: __.V
    ) -> tuple[ __.H, __.V ]:
        ''' Validates and/or prepares entry before addition.

            Should raise appropriate exception if entry is invalid.
        '''
        return key, value

    @__.abc.abstractmethod
    def _store_item_( self, key: __.H, value: __.V ) -> None:
        ''' Stores entry in underlying storage. '''
        raise NotImplementedError # pragma: no coverage

    def __setitem__( self, key: __.H, value: __.V ) -> None:
        key, value = self._pre_setitem_( key, value )
        if key in self:
            from .exceptions import EntryImmutabilityError
            raise EntryImmutabilityError( key )
        self._store_item_( key, value )

    def __delitem__( self, key: __.H ) -> None:
        from .exceptions import EntryImmutabilityError
        raise EntryImmutabilityError( key )

    def setdefault( self, key: __.H, default: __.V ) -> __.V:
        ''' Returns value for key, setting it to default if missing. '''
        try: return self[ key ]
        except KeyError:
            self[ key ] = default
            return default

    def update(
        self,
        *iterables: __.DictionaryPositionalArgument[ __.H, __.V ],
        **entries: __.DictionaryNominativeArgument[ __.V ],
    ) -> __.typx.Self:
        ''' Adds new entries as a batch. Returns self. '''
        from itertools import chain
        updates: list[ tuple[ __.H, __.V ] ] = [ ]
        for indicator, value in chain.from_iterable( map( # type: ignore
            lambda element: ( # type: ignore
                element.items( )
                if isinstance( element, __.cabc.Mapping )
                else element
            ),
            ( *iterables, entries )
        ) ):
            indicator_, value_ = (
                self._pre_setitem_( indicator, value ) ) # type: ignore
            if indicator_ in self:
                from .exceptions import EntryImmutabilityError
                raise EntryImmutabilityError( indicator_ )
            updates.append( ( indicator_, value_ ) )
        for indicator, value in updates: self._store_item_( indicator, value )
        return self


class _DictionaryOperations( AbstractDictionary[ __.H, __.V ] ):
    ''' Mix-in providing additional dictionary operations. '''

    def __init__(
        self, *posargs: __.typx.Any, **nomargs: __.typx.Any
    ) -> None:
        super( ).__init__( *posargs, **nomargs )

    def __or__( self, other: __.cabc.Mapping[ __.H, __.V ] ) -> __.typx.Self:
        if not isinstance( other, __.cabc.Mapping ): return NotImplemented
        conflicts = set( self.keys( ) ) & set( other.keys( ) )
        if conflicts:
            from .exceptions import EntryImmutabilityError
            raise EntryImmutabilityError( next( iter( conflicts ) ) )
        data = dict( self )
        data.update( other )
        return self.with_data( data )

    def __ror__( self, other: __.cabc.Mapping[ __.H, __.V ] ) -> __.typx.Self:
        if not isinstance( other, __.cabc.Mapping ): return NotImplemented
        return self | other

    def __and__(
        self,
        other: __.cabc.Set[ __.H ] | __.cabc.Mapping[ __.H, __.V ]
    ) -> __.typx.Self:
        if isinstance( other, __.cabc.Mapping ):
            return self.with_data( # pyright: ignore
                ( key, value ) for key, value in self.items( )
                if key in other and other[ key ] == value )
        if isinstance( other, ( __.cabc.Set, __.cabc.KeysView ) ):
            return self.with_data( # pyright: ignore
                ( key, self[ key ] ) for key in self.keys( ) & other )
        return NotImplemented

    def __rand__(
        self,
        other: __.cabc.Set[ __.H ] | __.cabc.Mapping[ __.H, __.V ]
    ) -> __.typx.Self:
        if not isinstance(
            other, ( __.cabc.Mapping, __.cabc.Set, __.cabc.KeysView )
        ): return NotImplemented
        return self & other

    @__.abc.abstractmethod
    def copy( self ) -> __.typx.Self:
        ''' Provides fresh copy of dictionary. '''
        raise NotImplementedError # pragma: no coverage

    @__.abc.abstractmethod
    def with_data(
        self,
        *iterables: __.DictionaryPositionalArgument[ __.H, __.V ],
        **entries: __.DictionaryNominativeArgument[ __.V ],
    ) -> __.typx.Self:
        ''' Creates new dictionary with same behavior but different data. '''
        raise NotImplementedError # pragma: no coverage


class _Dictionary(
    __.AccretiveDictionary[ __.H, __.V ], metaclass = _classes.Class
): pass


class Dictionary( # pylint: disable=eq-without-hash
    _DictionaryOperations[ __.H, __.V ]
):
    ''' Accretive dictionary. '''

    __slots__ = ( '_data_', )

    _data_: _Dictionary[ __.H, __.V ]

    def __init__(
        self,
        *iterables: __.DictionaryPositionalArgument[ __.H, __.V ],
        **entries: __.DictionaryNominativeArgument[ __.V ],
    ) -> None:
        self._data_ = _Dictionary( *iterables, **entries )
        super( ).__init__( )

    def __iter__( self ) -> __.cabc.Iterator[ __.H ]:
        return iter( self._data_ )

    def __len__( self ) -> int:
        return len( self._data_ )

    def __repr__( self ) -> str:
        return "{fqname}( {contents} )".format(
            fqname = __.calculate_fqname( self ),
            contents = str( self._data_ ) )

    def __str__( self ) -> str:
        return str( self._data_ )

    def __contains__( self, key: __.typx.Any ) -> bool:
        return key in self._data_

    def __getitem__( self, key: __.H ) -> __.V:
        return self._data_[ key ]

    def __eq__( self, other: __.typx.Any ) -> __.ComparisonResult:
        if isinstance( other, __.cabc.Mapping ):
            return self._data_ == other
        return NotImplemented

    def __ne__( self, other: __.typx.Any ) -> __.ComparisonResult:
        if isinstance( other, __.cabc.Mapping ):
            return self._data_ != other
        return NotImplemented

    def copy( self ) -> __.typx.Self:
        ''' Provides fresh copy of dictionary. '''
        return type( self )( self )

    def get(
        self, key: __.H, default: __.Absential[ __.V ] = __.absent
    ) -> __.typx.Annotated[
        __.V,
        __.typx.Doc(
            'Value of entry, if it exists. '
            'Else, supplied default value or ``None``.' )
    ]:
        ''' Retrieves entry associated with key, if it exists. '''
        if __.is_absent( default ):
            return self._data_.get( key ) # pyright: ignore
        return self._data_.get( key, default )

    def keys( self ) -> __.cabc.KeysView[ __.H ]:
        ''' Provides iterable view over dictionary keys. '''
        return self._data_.keys( )

    def items( self ) -> __.cabc.ItemsView[ __.H, __.V ]:
        ''' Provides iterable view over dictionary items. '''
        return self._data_.items( )

    def values( self ) -> __.cabc.ValuesView[ __.V ]:
        ''' Provides iterable view over dictionary values. '''
        return self._data_.values( )

    def with_data(
        self,
        *iterables: __.DictionaryPositionalArgument[ __.H, __.V ],
        **entries: __.DictionaryNominativeArgument[ __.V ],
    ) -> __.typx.Self:
        return type( self )( *iterables, **entries )

    def _store_item_( self, key: __.H, value: __.V ) -> None:
        self._data_[ key ] = value

Dictionary.__doc__ = __.generate_docstring(
    Dictionary, 'dictionary entries accretion' )


class ProducerDictionary( Dictionary[ __.H, __.V ] ):
    ''' Accretive dictionary with default value for missing entries. '''

    __slots__ = ( '_producer_', )

    _producer_: __.DictionaryProducer[ __.V ]

    def __init__(
        self,
        producer: __.DictionaryProducer[ __.V ],
        /,
        *iterables: __.DictionaryPositionalArgument[ __.H, __.V ],
        **entries: __.DictionaryNominativeArgument[ __.V ],
    ):
        # TODO: Validate producer argument.
        self._producer_ = producer
        super( ).__init__( *iterables, **entries )

    def __repr__( self ) -> str:
        return "{fqname}( {producer}, {contents} )".format(
            fqname = __.calculate_fqname( self ),
            producer = self._producer_,
            contents = str( self._data_ ) )

    def __getitem__( self, key: __.H ) -> __.V:
        if key not in self:
            value = self._producer_( )
            self[ key ] = value
        else: value = super( ).__getitem__( key )
        return value

    def copy( self ) -> __.typx.Self:
        ''' Provides fresh copy of dictionary. '''
        dictionary = type( self )( self._producer_ )
        return dictionary.update( self )

    def setdefault( self, key: __.H, default: __.V ) -> __.V:
        ''' Returns value for key, setting it to default if missing. '''
        if key not in self:
            self[ key ] = default
            return default
        return self[ key ]

    def with_data(
        self,
        *iterables: __.DictionaryPositionalArgument[ __.H, __.V ],
        **entries: __.DictionaryNominativeArgument[ __.V ],
    ) -> __.typx.Self:
        return type( self )( self._producer_, *iterables, **entries )

ProducerDictionary.__doc__ = __.generate_docstring(
    ProducerDictionary,
    'dictionary entries accretion',
    'dictionary entries production',
)


class ValidatorDictionary( Dictionary[ __.H, __.V ] ):
    ''' Accretive dictionary with validation of new entries. '''

    __slots__ = ( '_validator_', )

    _validator_: __.DictionaryValidator[ __.H, __.V ]

    def __init__(
        self,
        validator: __.DictionaryValidator[ __.H, __.V ],
        /,
        *iterables: __.DictionaryPositionalArgument[ __.H, __.V ],
        **entries: __.DictionaryNominativeArgument[ __.V ],
    ) -> None:
        self._validator_ = validator
        super( ).__init__( *iterables, **entries )

    def __repr__( self ) -> str:
        return "{fqname}( {validator}, {contents} )".format(
            fqname = __.calculate_fqname( self ),
            validator = self._validator_,
            contents = str( self._data_ ) )

    def _pre_setitem_( self, key: __.H, value: __.V ) -> tuple[ __.H, __.V ]:
        if not self._validator_( key, value ):
            from .exceptions import EntryValidityError
            raise EntryValidityError( key, value )
        return key, value

    def copy( self ) -> __.typx.Self:
        ''' Provides fresh copy of dictionary. '''
        dictionary = type( self )( self._validator_ )
        return dictionary.update( self )

    def with_data(
        self,
        *iterables: __.DictionaryPositionalArgument[ __.H, __.V ],
        **entries: __.DictionaryNominativeArgument[ __.V ],
    ) -> __.typx.Self:
        return type( self )( self._validator_, *iterables, **entries )

ValidatorDictionary.__doc__ = __.generate_docstring(
    ValidatorDictionary,
    'dictionary entries accretion',
    'dictionary entries validation',
)


class ProducerValidatorDictionary( Dictionary[ __.H, __.V ] ):
    ''' Accretive dictionary with defaults and validation. '''

    __slots__ = ( '_producer_', '_validator_' )

    _producer_: __.DictionaryProducer[ __.V ]
    _validator_: __.DictionaryValidator[ __.H, __.V ]

    def __init__(
        self,
        producer: __.DictionaryProducer[ __.V ],
        validator: __.DictionaryValidator[ __.H, __.V ],
        /,
        *iterables: __.DictionaryPositionalArgument[ __.H, __.V ],
        **entries: __.DictionaryNominativeArgument[ __.V ],
    ) -> None:
        self._producer_ = producer
        self._validator_ = validator
        super( ).__init__( *iterables, **entries )

    def __repr__( self ) -> str:
        return "{fqname}( {producer}, {validator}, {contents} )".format(
            fqname = __.calculate_fqname( self ),
            producer = self._producer_,
            validator = self._validator_,
            contents = str( self._data_ ) )

    def __getitem__( self, key: __.H ) -> __.V:
        if key not in self:
            value = self._producer_( )
            if not self._validator_( key, value ):
                from .exceptions import EntryValidityError
                raise EntryValidityError( key, value )
            self[ key ] = value
        else: value = super( ).__getitem__( key )
        return value

    def _pre_setitem_( self, key: __.H, value: __.V ) -> tuple[ __.H, __.V ]:
        if not self._validator_( key, value ):
            from .exceptions import EntryValidityError
            raise EntryValidityError( key, value )
        return key, value

    def copy( self ) -> __.typx.Self:
        ''' Provides fresh copy of dictionary. '''
        dictionary = type( self )( self._producer_, self._validator_ )
        return dictionary.update( self )

    def setdefault( self, key: __.H, default: __.V ) -> __.V:
        ''' Returns value for key, setting it to default if missing. '''
        if key not in self:
            self[ key ] = default
            return default
        return self[ key ]

    def with_data(
        self,
        *iterables: __.DictionaryPositionalArgument[ __.H, __.V ],
        **entries: __.DictionaryNominativeArgument[ __.V ],
    ) -> __.typx.Self:
        return type( self )(
            self._producer_, self._validator_, *iterables, **entries )

ProducerValidatorDictionary.__doc__ = __.generate_docstring(
    ProducerValidatorDictionary,
    'dictionary entries accretion',
    'dictionary entries production',
    'dictionary entries validation',
)



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Docstring utilities. '''

# pylint: disable=unused-wildcard-import,wildcard-import
# ruff: noqa: F403,F405


from __future__ import annotations

from . import doctab
from .imports import *


class Docstring( str ):
    ''' Dedicated docstring container. '''


def generate_docstring(
    *fragment_ids: type | Docstring | str,
    table: cabc.Mapping[ str, str ] = doctab.TABLE,
) -> str:
    ''' Sews together docstring fragments into clean docstring. '''
    from inspect import cleandoc, getdoc, isclass
    fragments: list[ str ] = [ ]
    for fragment_id in fragment_ids:
        if isclass( fragment_id ): fragment = getdoc( fragment_id ) or ''
        elif isinstance( fragment_id, Docstring ): fragment = fragment_id
        else: fragment = table[ fragment_id ]
        fragments.append( cleandoc( fragment ) )
    return '\n\n'.join( fragments )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Docstrings table for reuse across entities. '''

# pylint: disable=unused-wildcard-import,wildcard-import
# ruff: noqa: F403,F405


from __future__ import annotations

from .imports import *


TABLE: types.MappingProxyType[ str, str ] = types.MappingProxyType( {

    'class attributes accretion': '''
Prevents reassignment or deletion of class attributes after they have been
assigned. Only assignment of new class attributes is permitted.
''',

    'description of class factory class': '''
Derived from :py:class:`type`, this is a metaclass. A metaclass is a class
factory class. I.e., it is a class that produces other classes as its
instances.
''',

    'description of module': '''
Derived from :py:class:`types.ModuleType`, this class is suitable for use as a
Python module class.
''',

    'description of namespace': '''
A namespace is an object, whose attributes can be determined from iterables and
keyword arguments, at initialization time. The string representation of the
namespace object reflects its current instance attributes. Modeled after
:py:class:`types.SimpleNamespace`.
''',

    'dictionary entries accretion': '''
Prevents alteration or removal of dictionary entries after they have been
added. Only addition of new dictionary entries is permitted.
''',

    'dictionary entries production': '''
When an attempt to access a missing entry is made, then the entry is added with
a default value. Modeled after :py:class:`collections.defaultdict`.
''',

    'dictionary entries validation': '''
When an attempt to add a new entry is made, then the entry is validated against
supplied criteria. If validation fails, then the entry is rejected.
''',

    'instance attributes accretion': '''
Prevents reassignment or deletion of instance attributes after they have been
assigned. Only assignment of new instance attributes is permitted.
''',

    'module attributes accretion': '''
Prevents reassignment or deletion of module attributes after they have been
assigned. Only assignment of new module attributes is permitted.

This behavior helps ensure that module-level constants remain constant and that
module interfaces remain stable during runtime.
''',

} )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Qualified aliases to accretive data structures.

    Provides aliases prefixed with "Accretive" for all core classes. These are
    useful for avoiding namespace collisions when importing from the package,
    particularly with common names like "Dictionary" or "Namespace".

    For example, instead of:

    >>> from accretive import Dictionary
    >>> # Possible conflict with other Dictionary classes

    you could use:

    >>> from accretive.qaliases import AccretiveDictionary
    >>> # Clearly indicates the source and behavior
'''


# ruff: noqa: F401
# pylint: disable=unused-import


from . import __
from .classes import (
    ABCFactory as           AccretiveABCFactory,
    Class as                AccretiveClass,
    ProtocolClass as        AccretiveProtocolClass,
)
from .dictionaries import (
    AbstractDictionary as           AbstractAccretiveDictionary,
    Dictionary as                   AccretiveDictionary,
    ProducerDictionary as           AccretiveProducerDictionary,
    ProducerValidatorDictionary as  AccretiveProducerValidatorDictionary,
    ValidatorDictionary as          AccretiveValidatorDictionary,
)
from .modules import (
    Module as               AccretiveModule,
    reclassify_modules as   reclassify_modules_as_accretive,
)
from .namespaces import (
    Namespace as            AccretiveNamespace,
)
from .objects import (
    Object as               AccretiveObject,
                            accretive,
)



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Attribute concealment and immutability. '''


from __future__ import annotations

import collections.abc as cabc
import types

import typing_extensions as typx


ClassDecorators: typx.TypeAlias = (
    cabc.Iterable[ cabc.Callable[ [ type ], type ] ] )


behavior_label = 'immutability'


def repair_class_reproduction( original: type, reproduction: type ) -> None:
    ''' Repairs a class reproduction, if necessary. '''
    from platform import python_implementation
    match python_implementation( ):
        case 'CPython':  # pragma: no branch
            _repair_cpython_class_closures( original, reproduction )
        case _: pass  # pragma: no cover


def _repair_cpython_class_closures( # pylint: disable=too-complex
    original: type, reproduction: type
) -> None:
    # Adapted from https://github.com/python/cpython/pull/124455/files
    def try_repair_closure( function: cabc.Callable[ ..., typx.Any ] ) -> bool:
        try: index = function.__code__.co_freevars.index( '__class__' )
        except ValueError: return False
        if not function.__closure__: return False # pragma: no branch
        closure = function.__closure__[ index ]
        if original is closure.cell_contents: # pragma: no branch
            closure.cell_contents = reproduction
            return True
        return False # pragma: no cover

    from inspect import isfunction, unwrap
    for attribute in reproduction.__dict__.values( ): # pylint: disable=too-many-nested-blocks
        attribute_ = unwrap( attribute )
        if isfunction( attribute_ ) and try_repair_closure( attribute_ ):
            return
        if isinstance( attribute_, property ):
            for aname in ( 'fget', 'fset', 'fdel' ):
                accessor = getattr( attribute_, aname )
                if None is accessor: continue
                if try_repair_closure( accessor ): return # pragma: no branch


class ImmutableClass( type ):
    ''' Concealment and immutability on class attributes. '''

    _class_attribute_visibility_includes_: cabc.Collection[ str ] = (
        frozenset( ) )

    def __new__(
        clscls: type[ type ],
        name: str,
        bases: tuple[ type, ... ],
        namespace: dict[ str, typx.Any ], *,
        decorators: ClassDecorators = ( ),
        **args: typx.Any
    ) -> ImmutableClass:
        class_ = type.__new__( clscls, name, bases, namespace, **args )
        return _immutable_class__new__( class_, decorators = decorators )

    def __init__( selfclass, *posargs: typx.Any, **nomargs: typx.Any ):
        super( ).__init__( *posargs, **nomargs )
        _immutable_class__init__( selfclass )

    def __dir__( selfclass ) -> tuple[ str, ... ]:
        default: frozenset[ str ] = frozenset( )
        includes: frozenset[ str ] = frozenset.union( *( # type: ignore
            getattr( class_, '_class_attribute_visibility_includes_', default )
            for class_ in selfclass.__mro__ ) )
        return tuple( sorted(
            name for name in super( ).__dir__( )
            if not name.startswith( '_' ) or name in includes ) )

    def __delattr__( selfclass, name: str ) -> None:
        if not _immutable_class__delattr__( selfclass, name ):
            super( ).__delattr__( name )

    def __setattr__( selfclass, name: str, value: typx.Any ) -> None:
        if not _immutable_class__setattr__( selfclass, name ):
            super( ).__setattr__( name, value )


def _immutable_class__new__(
    original: type,
    decorators: ClassDecorators = ( ),
) -> type:
    # Some decorators create new classes, which invokes this method again.
    # Short-circuit to prevent recursive decoration and other tangles.
    decorators_ = original.__dict__.get( '_class_decorators_', [ ] )
    if decorators_: return original
    setattr( original, '_class_decorators_', decorators_ )
    reproduction = original
    for decorator in decorators:
        decorators_.append( decorator )
        reproduction = decorator( original )
        if original is not reproduction:
            repair_class_reproduction( original, reproduction )
        original = reproduction
    decorators_.clear( )  # Flag '__init__' to enable immutability
    return reproduction


def _immutable_class__init__( class_: type ) -> None:
    # Some metaclasses add class attributes in '__init__' method.
    # So, we wait until last possible moment to set immutability.
    if class_.__dict__.get( '_class_decorators_' ): return
    del class_._class_decorators_
    if ( class_behaviors := class_.__dict__.get( '_class_behaviors_' ) ):
        class_behaviors.add( behavior_label )
    else: setattr( class_, '_class_behaviors_', { behavior_label } )


def _immutable_class__delattr__( class_: type, name: str ) -> bool:
    # Consult class attributes dictionary to ignore immutable base classes.
    if behavior_label not in class_.__dict__.get(
        '_class_behaviors_', ( )
    ): return False
    raise AttributeError(
        "Cannot delete attribute {name!r} "
        "on class {class_fqname!r}.".format(
            name = name,
            class_fqname = calculate_class_fqname( class_ ) ) )


def _immutable_class__setattr__( class_: type, name: str ) -> bool:
    # Consult class attributes dictionary to ignore immutable base classes.
    if behavior_label not in class_.__dict__.get(
        '_class_behaviors_', ( )
    ): return False
    raise AttributeError(
        "Cannot assign attribute {name!r} "
        "on class {class_fqname!r}.".format(
            name = name,
            class_fqname = calculate_class_fqname( class_ ) ) )


class ConcealerExtension:
    ''' Conceals instance attributes according to some criteria.

        By default, public attributes are displayed.
    '''

    _attribute_visibility_includes_: cabc.Collection[ str ] = frozenset( )

    def __dir__( self ) -> tuple[ str, ... ]:
        return tuple( sorted(
            name for name in super( ).__dir__( )
            if not name.startswith( '_' )
                or name in self._attribute_visibility_includes_ ) )


class ImmutableModule(
    ConcealerExtension, types.ModuleType, metaclass = ImmutableClass
):
    ''' Concealment and immutability on module attributes. '''

    def __delattr__( self, name: str ) -> None:
        raise AttributeError( # noqa: TRY003
            f"Cannot delete attribute {name!r} "
            f"on module {self.__name__!r}." ) # pylint: disable=no-member

    def __setattr__( self, name: str, value: typx.Any ) -> None:
        raise AttributeError( # noqa: TRY003
            f"Cannot assign attribute {name!r} "
            f"on module {self.__name__!r}." ) # pylint: disable=no-member


class ImmutableObject( ConcealerExtension, metaclass = ImmutableClass ):
    ''' Concealment and immutability on instance attributes. '''

    def __delattr__( self, name: str ) -> None:
        raise AttributeError(
            "Cannot delete attribute {name!r} on instance "
            "of class {class_fqname!r}.".format(
                name = name, class_fqname = calculate_fqname( self ) ) )

    def __setattr__( self, name: str, value: typx.Any ) -> None:
        raise AttributeError(
            "Cannot assign attribute {name!r} on instance "
            "of class {class_fqname!r}.".format(
                name = name, class_fqname = calculate_fqname( self ) ) )


def calculate_class_fqname( class_: type ) -> str:
    ''' Calculates fully-qualified name for class. '''
    return f"{class_.__module__}.{class_.__qualname__}"


def calculate_fqname( obj: object ) -> str:
    ''' Calculates fully-qualified name for class of object. '''
    class_ = type( obj )
    return f"{class_.__module__}.{class_.__qualname__}"


def discover_public_attributes(
    attributes: cabc.Mapping[ str, typx.Any ]
) -> tuple[ str, ... ]:
    ''' Discovers public attributes of certain types from dictionary.

        By default, callables, including classes, are discovered.
    '''
    return tuple( sorted(
        name for name, attribute in attributes.items( )
        if not name.startswith( '_' ) and callable( attribute ) ) )

def reclassify_modules(
    attributes: typx.Annotated[
        cabc.Mapping[ str, typx.Any ] | types.ModuleType | str,
        typx.Doc( 'Module, module name, or dictionary of object attributes.' ),
    ],
    recursive: typx.Annotated[
        bool,
        typx.Doc( 'Recursively reclassify package modules?' ),
    ] = False,
) -> None:
    ''' Reclassifies modules to be immutable. '''
    from inspect import ismodule
    from sys import modules
    if isinstance( attributes, str ):
        attributes = modules[ attributes ]
    if isinstance( attributes, types.ModuleType ):
        module = attributes
        attributes = attributes.__dict__
    else: module = None
    package_name = (
        attributes.get( '__package__' ) or attributes.get( '__name__' ) )
    if not package_name: return
    for value in attributes.values( ):
        if not ismodule( value ): continue
        if not value.__name__.startswith( f"{package_name}." ): continue
        if recursive: reclassify_modules( value, recursive = True )
        if isinstance( value, ImmutableModule ): continue
        value.__class__ = ImmutableModule
    if module and not isinstance( module, ImmutableModule ):
        module.__class__ = ImmutableModule
